#ifndef INCLUDED_DEFS_H
#define INCLUDED_DEFS_H

typedef double scalar;
typedef scalar *vector;
typedef vector *matrix;
/* an element of a symmetric matrix created by symatrixalloc */
/* notice that SYMELEM gives a value, and therefore must me a rhs */
#define SYMELEM(matrix, i, j) ((i)<=(j) ? m[i][j] : m[j][i])

/* types of matrices or vectors */
#define M_SYMATRIX	0	/* symmetric real matrix */
#define M_TRIDIAG	1	/* symmetric real tridiagonal matrix */
#define M_VECTOR	2	/* vector */
#define M_GMATRIX	3	/* general square matrix */
#define M_RMATRIX	4	/* a rectangular matrix */

/* sturm related types */
typedef struct Sturm Sturm;
typedef struct trid trid;

struct trid {
	vector a,b;
	int length;
};

struct Sturm {		/* linked list of sturm data for each */
	trid t;			/* of the irreducible matrices in the given */
				/* matrix */
	int flag;		/* for use by findroots */
	struct {
		scalar upper;
		scalar lower;
	} bound;
	Sturm *next;
};
#endif

#ifndef INCLUDED_PORI_H
#define INCLUDED_PORI_H

/* from memory.c */
extern matrix	symatrixalloc(), tridiagalloc(), matrixalloc(),rectalloc();
extern vector	vectoralloc();
extern		symatrixfree(), tridiagfree(), vectorfree(), matrixfree();
/* from strum.c */
extern Sturm	*sturm_list;
extern scalar	findroot();
extern int	sign_changes(), tsolve();
extern		sturm_gen(), findbounds(), findroots();
/* from tridiag.c */
extern		householder();
/* from eigen.c */
extern		eigen();
/* from etc.c */
extern		randomvec(), vvsub();
extern scalar	norm(), normalize();

/* from stats.c */
#ifdef STATS
extern long	s_findroot,s_signchanges,s_schangeslen,s_harel,s_harellen,
		s_tsolve,s_tsolvelen;
extern		statsprint();
#endif

/* from standard libraries */
extern double	sqrt();

/* useful defines */

#define ABS(x)	((x)>0 ? (x) : (-(x)))
#define MIN(a,b)	((a)<(b) ? (a) : (b))
#define EPSILON 1e-15
/*#define STURMEPSILON 1e-6*/
#define MAXITER 100

#define POWERSMOOTH	0
#define ADAPTIVESMOOTH	1

#endif
