#include "eigen.h"
#include "stdio.h"

char *malloc();

static memerror(size, type, size2)
{
	fprintf(stderr, "Internal error in memory.c:\n");
	fprintf(stderr, "\tNot enough memory for object.\n");	
	fprintf(stderr, "Object type to be allocated: ");
	switch(type){
	case M_SYMATRIX:	fprintf(stderr,
		"%dx%d real symmetric matrix\n",size,size);
		break;
	case M_TRIDIAG:	fprintf(stderr,
		"%dx%d real symmetric tridiagonal\n",size,size);
		break;
	case M_VECTOR:	fprintf(stderr, "%d vector\n",size);
		break;
	case M_GMATRIX:	fprintf(stderr, "%dx%d real matrix\n",size,size);
		break;
	case M_RMATRIX: fprintf(stderr, "%dx%d real matrix\n",size,size2);
		break;
	}
	exit(1);
	
}

/* symatrixalloc(size) - allocate a sizeXsize symetric matrix,
			 by allocating the lower triangle
*/
matrix
symatrixalloc(size)
	int size;
{
	matrix m;
	int s;

	if(!(m=(matrix)malloc(size*(sizeof(vector)))))
		memerror(size,M_SYMATRIX,0);

	s=size;
	for(s=size;--s;){
		if(!(m[s-1]=(vector)malloc(s*(sizeof(scalar)))))
			memerror(size,M_SYMATRIX,0);
	}
	return m;
}

/*symatrixfree(matrix, size) - deallocate a matrix allocated by symatrixalloc*/
symatrixfree(m,size)
	matrix m;
	int size;
{
	while(size--) free(m[size]);
	free(m);
}
/* matrixalloc(size) -	allocate a sizeXsize real matrix,
*/
matrix
matrixalloc(size)
	int size;
{
	matrix m;
	int i;

	if(!(m=(matrix)malloc(size*(sizeof(vector)))))
		memerror(size,M_GMATRIX,0);

	for(i=0;i<size;i++){
		if(!(m[i]=(vector)malloc(size*(sizeof(scalar)))))
			memerror(size,M_GMATRIX,0);
	}
	return m;
}

/*matrixfree(matrix, size) - deallocate a matrix allocated by matrixalloc*/
matrixfree(m,size)
	matrix m;
	int size;
{
	while(size--) free(m[size]);
	free(m);
}
/* rectalloc(rows,cols) -	allocate a rowsXcols real matrix,
   to deacllocate, call matrixfree with rows instead of size.
*/
matrix
rectalloc(rows,cols)
	int rows,cols;
{
	matrix m;
	int i;

	if(!(m=(matrix)malloc(rows*(sizeof(vector)))))
		memerror(rows,M_RMATRIX,cols);

	for(i=0;i<rows;i++){
		if(!(m[i]=(vector)malloc(cols*(sizeof(scalar)))))
			memerror(rows,M_RMATRIX,cols);
	}
	return m;
}

/*tridiagalloc(size) - allocate a NxN symmetric tridiagonal matrix */
/* note that before the second vector of a tridiagonal matrix there
   is an extra allocated scalar, for temporary use in householder.
   note that m[1][0] is just as you expect. m[1][-1] is the extra
   scalar which is only used by householder
*/
matrix
tridiagalloc(size)
	int size;
{
	matrix m;
	if(!(m=(matrix)malloc(2*sizeof(vector))))
		memerror(size,M_TRIDIAG,0);
	if(!(m[0]=(vector)malloc(size*sizeof(scalar))))
		memerror(size,M_TRIDIAG,0);
	if(!(m[1]=(vector)malloc(size*sizeof(scalar))+1))
		memerror(size,M_TRIDIAG,0);
	return m;
}

tridiagfree(m)
	matrix m;
{
	free(m[0]);
	free(m[1]-1);
	free(m);
}

vector
vectoralloc(size)
{
	vector v;
	if(!(v=(vector)malloc(size*sizeof(scalar))))
		memerror(size,M_VECTOR,0);
	return v;
}

vectorfree(v)
	vector v;
{
	free(v);
}
