.AM
.EQ
delim $$
define sign ' roman "sign" ^ '
define if ' ~ roman "if" ~ '
.EN
.TL
Finding the largest eigenvalues of a real symmetric matrix, and corresponding eigenvectors
.AU
Nadav Har'El
.AI
Department of Mathematics
Technion \- Israel Institute of Technology
Haifa 32000, Israel
E-Mail: nyh@gauss.technion.ac.il
.AB
This report describes my solution to the problem of finding the largest
eigenvalues of a real symmetric matrix, and their corresponding eigenvectors,
without finding all the eigenvalues of the given matrix.
.AE
.NH
Introduction.
.PP
Many problems require finding only a small part of the eigenvalues and
eigenvectors of a large real symmetric matrix. For such problems, finding
all the eigenvectors and eigenvalues, with algorithms such as Jacobi, or
the QL method (cf. [NRC]), may be time consuming and wasteful. An example
of such a problem is the Karhunen-Loeve expansion in pattern recognition:
Given an ensemble of images, let say faces, one wants to find an optimal
base for the vector space of images of the given size. The base has to be
optimal in the sense that when an image similar to the ones in the ensemble
(such as another face) is represented in that base, we can keep only, say,
the first 40 base elements, and drop the others, while the error induced by
that action is relatively small. The optimal base also has to be orthonormal,
so representing an image in that base can be easily done. To find that optimal
base, one has to define a certain real symmetric matrix, whose eigenvectors
are the required base, where the most important vectors are the ones corresponding
to the largest eigenvalues, and the ones corresponding to small eigenvalues can be
dropped.
For a more complete explanation of this problem see [KL1] and [KL2].
Another problem requiring only part of the eigenvalues and eigenvectors has
to do with solving differential equations, which I will not get into in this
report.
.NH
The basic idea of the algorithm
.PP
The basic idea of the algorithm is as follows: First we convert the given
symmetric matrix to a similar tridiagonal matrix (i.e. has nonzero elements
only on the diagonal and sub-diagonals), using the 
.I Householder
algorithm (see below). We also get the appropriate transformation matrix.
Now that we have reduced our problem to real symmetric tridiagonal matrices,
we use a 
.I "Sturm Sequences"
based algorithm (see below) to find the $N$ largest eigenvalues, and their
corresponding (orthonormal) eigenvectors, of the tridiagonal matrix. Finally, we use
the transformation matrix to convert these eigenvectors to the ones of
the original matrix. The orthonormality of the eigenvectors are preserved by
this conversion, since the transformation matrix is orthogonal.
.NH
The Householder algorithm
.PP
This section describes the Householder method of reduction of a symmetric matrix to tridiagonal form,
as defined in [NRC]. The following section is an extract from [NRC], pages 368-372.
.PP
The Householder algorithm reduces an $n times n$ symmetric matrix A to
tridiagonal form by $n - 2$ orthogonal transformations.
Each transformation annihilates the required part of a whole column and whole
corresponding row. The basic ingredient is a Householder matrix $P$, which has
the form
.EQ (1)
P = I - 2 w w sup T
.EN
where $w$ is a real vector with $| w | sup 2 = 1$. (In the present notation,
the
.I outer
or matrix product of two vectors, $a$ and $b$ is written $a b sup T$, while
the
.I inner
or scalar product of the vectors is written as $a sup T b$.) The matrix $P$ is
orthogonal, because
.EQ (2)
P sup 2 = ( 1 - 2 w w sup T ) ( 1 - 2 w w sup T)
= 1 - 4 w w sup T + 4 w ( w sup T w ) w sup T
= 1
.EN 
Therefore $P = P sup { -1 }$. But $P sup T = P$, and so $P sup T = P sup { -1 }$,
proving orthogonality.
.PP
Rewrite $P$ as
.EQ (3)
P = 1 - { u u sup T } over H ,
.EN
where the scalar H is
.EQ (4)
H == {1 over 2} | u | sup 2
.EN
and $u$ can now be any vector. Suppose $x$ is the vector composed of the first
column of A. Choose
.EQ (5)
u = x +- | x | e sub 1
.EN
where $e sub 1$ is the unit vector $ ( 1 , 0 , ... , 0 ) sup T$, and the choice
of signs will be made later. Then
.EQ (6)
P x = x - {u over H} (x +- | x | e sub 1 ) sup T x
= x - " " {2 u ( | x | sup 2 +- | x | x sub 1 )} over { 2 |x| sup 2 +- 2 | x | x sub 1}
" " = x - u
= - +- e sub 1
.EN
This shows that the Householder matrix $P$ acts on a given vector x to zero
all its elements except the first one.
.PP
To reduce a symmetric matrix A to tridiagonal form, we choose the vector x for
the first Householder matrix to be the lower $n - 1$ elements of the first
column. Then the the lower $n-2$ elements will be zeroed:
.EQ I
P sub 1 A mark = left [ matrix { ccol { 1 above 0 above 0 above ... above 0 }
ccol {0} ccol {0 above "" above {"" sup {(n-1)} P sub 1} } ccol {...}  ccol {0} } right ] ~ cdot ~
left [ matrix { ccol { { a sub {1 1} } above { a sub {2 1} } above { a sub {3 1} } above ... above { a sub {n 1} } } ccol {a sub {1 2} } ccol { {a sub {1 3}} above "" above "*"} ccol {...} ccol {a sub {1 n}} } right ]
.EN
.EQ I (7)
lineup = left [ matrix { ccol { { a sub {1 1} } above k above 0 above ... above 0 }
ccol {{a sub {1 2}}} ccol { {a sub {1 3}} above "" above "*" } ccol {...}
ccol {{a sub {1 n}}}} right ]
.EN
.PP
Here we have written the matrices in partitioned form, with $"" sup (n-1) P$
denoting a Householder matrix with dimensions $( n - 1 ) times ( n - 1)$,
and $"*"$ denoting an irrelevant part of the matrix.
The quantity $k$ is simply plus or minus the magnitude of the vector
$( a sub {2 1} ,..., a sub {n 1} ) sup T$.
.PP
The complete orthogonal transformation is now
.EQ (8)
A prime = P A P = 
left [ matrix { ccol { { a sub {1 1} } above k above 0 above ... above 0 }
ccol {k} ccol { 0 above "" above "*" } ccol {...}
ccol {0 }} right ]
.EN
.PP
Now choose the vector $x$ for the second Householder matrix to be the bottom
$n - 2$ elements of the second column, and from it construct
.EQ (9)
P sub 2 ==
left [ matrix { ccol { 1 above 0 above 0 above ... above 0 }
ccol {0 above 1 above 0 above ... above 0}
ccol {0 above 0 }
ccol {... above ... above "" above {"" sup {(n-2)} P sub 2} }
ccol {0 above 0}
} right ]
.EN
The identity block in the upper left corner of $P sub 2$ insures that the
tridiagonalization achieved in the first step will not be spoiled by this one
while the $(n - 2)$-dimensional Householder matrix at the lower corner of $P
sub 2$ creates one additional row and column of the tridiagonal output.
Clearly, a sequence of $n - 2$ such transformations will reduce the matrix
A to tridiagonal form.
.PP
Instead of actually carrying out the matrix multiplications in $P A P$, we
compute a vector
.EQ (10)
p == { A u } over H
.EN
Then
.EQ
A P = A (1 - { u u sup T } over H ) = A - p u sup T
.EN
.EQ
A prime = P A P = A - p u sup T - u p sup T + 2 K u u sup T
.EN
where the scalar K is defined by
.EQ (11)
K = { u sup T p } over { 2 H }
.EN
.PP
If we write
.EQ (12)
q == p - K u
.EN
then we have
.EQ (13)
A prime = A - q u sup T - u q sup T
.EN
This is the computationally useful formula.
.PP
Following Wilkinson and Reinsch, the routine for Householder reduction used
by the algorithm described in this document actually starts in the $n sup th$
column of A, not the first as in the explanation above. In detail, the
equations are as follows: At stage $m$ ($m = 1, 2, ..., n-2$) the vector $u$
has the form
.EQ (14)
u sup T = (a sub {i 1} , ..., a sub {i , i - 2 } , a sub {i , i - 1} +- sqrt sigma , 0, ..., 0 )
.EN
Here
.EQ (15)
i == n - m + 1 = n , n - 1 ,..., 3
.EN
and the quantity $sigma$ ($| x | sup 2$ in our earlier notation) is
.EQ (16)
sigma = ( a sub {i 1} ) sup 2 + ... + (a sub {i, i - 1} ) sup 2
.EN
We choose the sign of $sigma$ in (9) to be the same as the sign of $a sub {i, i - 1 }$ to lessen roundoff error.
.PP
Variables are thus computed in the following order: $sigma , u , H , p , K, q , A prime$. At any stage $m$, $A$ is tridiagonal in it's last m-1 rows and columns.
.PP
If the eigenvectors of the final tridiagonal matrix are found, then the
eigenvectors of $A$ can be obtained by applying the accumulated transformation
.EQ (17)
Q = P sub 1 P sub 2 ... P sub { n - 2 }
.EN
to those eigenvectors. We therefore form $Q$ by recursion after all the $P$'s
have been determined:
.EQ
Q sub {n - 2} = P sub {n - 2}
.EN
.EQ (18)
Q sub j = P sub j Q sub { j + 1 } , j = n - 3 ,..., 1,
.EN
.EQ
Q = Q sub 1
.EN
.PP
It has been shown that the Householder algorithm described above, as implemented
In the routine I use, has the complexity of about $ 4 / 3 n sup 3 $ at the
limit of large $n$.
.NH
The Sturm sequence algorithm
.PP
The following section describes the algorithm of finding the greatest eigenvalues
of an irreducible real symmetric tridiagonal matrix. An irreducible tridiagonal matrix is a
tridiagonal matrix with no zeros on the subdiagonal. See [INA], page 281 for
farther discussion of Sturm Sequences and Bisection Methods.
.PP
Let $p ( x )$ be a polynomial of degree $n$,
.EQ
p ( x ) = a sub 0 x sup n + a sub 1 x sup {n - 1} + ... + a sub n^ ,^ a sub 0 != 0.
.EN
It is possible to determine the number of real roots of $p ( x )$ in a specified
region by examining the number of sign changes $w ( a )$ for certain points
$x = a$ of a sequence of polynomials $p sub i ( x ) ^ , ^ i = 0, 1, ..., m ,$ of
descending degrees. Such a sign change happens whenever the sign of a
polynomial value differs from that of its successor. Furthermore, if $p sub i
( a ) = 0$, then this entry is to be removed from the sequence of polynomial
values before the sign changes are counted. Suitable sequences of polynomials
are so called 
.I "Sturm Sequences."
.br
.B Definition.
The sequence
.EQ
p ( x ) = p sub 0 ( x ) , p sub 1 ( x ) , ..., p sub m ( x )
.EN
of real polynomials is a Sturm sequence for the polynomial $p ( x )$ if:
.br
.IP (a)
All real roots of $p sub 0 ( x )$ are simple.
.IP (b)
$sign  p sub 1 ( zeta ) = - sign  {p sub 0} prime ( zeta )$ if $zeta$ is a
real root of $p sub 0 ( x )$.
.IP (c)
For $i = 1, 2, ..., m - 1, ~ p sub {i + 1} ( zeta ) p sub {i - 1} ( zeta ) < 0$
if $zeta$ is a real root of $p sub i ( x )$.
.IP (d)
The last polynomial $p sub m ( x )$ has no real roots.
.LP
For such Sturm sequences we have the following
.br
.B Theorem.
The number of real roots of $p ( x ) == p sub 0 ( x )$ in the interval 
$a <= x < b$ equals $w ( b ) - w ( a )$, where $w ( x )$ is the number of sign
changes of a Sturm sequence
.EQ
p sub 0 ( x ), ..., p sub m ( x )
.EN
at location x.
.br
See proof at [INA].
.PP
An important use of Sturm sequences, and indeed what I use in the algorithm
I present in this report, is in
.I "bisection methods
for determining the eigenvalues of real symmetric matrices which are tridiagonal.
Recall the characteristic polynomials $p sub i ( x )$ of the principal minor
formed by the first i rows and columns of the matrix $( J - x I )$, which, as can be
seen, satisfy the recursion
.EQ (1)
p sub 0 ( x ) == 1 ,
.EN
.EQ
p sub 1 ( x ) == alpha sub 1 - x ,
.EN
.EQ
p sub i ( x ) == ( alpha sub i - x) p sub {i - 1} (x) - {beta sub i} sup 2
p sub {i - 2} (x), "    " i=2, 3, ..., n.
.EN
Where $alpha sub 1 , ..., alpha sub n$ are the diagonal of J, and
$beta sub 2 , ..., beta sub n$ are the subdiagonal.
The key observation is that the polynomials
.EQ
p sub n ( x ) , p sub {n - 1} ( x ) , ..., p sub 0 ( x )
.EN
are a Sturm sequence for the characteristic polynomial $p sub n ( x )
= det ( J - x I )$
(note that the polynomials here are indexed in the opposite order from before,
since this indexing seems more logical for a sequence generated by iteration),
provided the off-diagonal elements $beta sub i , ^ i = 2 ,..., n$,
of the tridiagonal matrix J are all nonzero. This is proven in [INA].
As a consequence of that proof we derive that $p sub n ( x )$ has simple real
roots $zeta sub 1 > zeta sub 2 > ... > zeta sub n$, and that
.EQ
sign  p sub {n - 1} ( zeta sub k ) = ( -1 ) sup { n + k } ,
.EN
.EQ
sign  {p sub n} prime ( zeta sub k ) = ( -1 ) sup {n + k + 1} = - sign   p sub {n - 1} ( zeta sub k ),
.EN
for k = 1, 2 ,..., n.
.PP
For $ x = - inf $ the Sturm sequence above has the sign pattern
.EQ
+ , + , ..., +.
.EN
Thus $w ( - inf ) = 0$. By the previous theorem, $w ( eta ) $ indicates the number
of roots $zeta$ of $p sub n ( x )$ with $zeta < eta$: $w ( eta ) >= n + 1 - i$
holds if and only if $zeta sub i < eta$.
.PP
The bisection method for determining the $i sup th$ root $zeta sub i$ of
$p sub n ( x )$ ($zeta sub 1 > zeta sub 2 > ... > zeta sub n$) now is as follows.
Start with an interval [$a sub 0 , b sub 0 $] which is known to contain 
$zeta sub 1$; e.g., choose $b sub 0 > zeta sub 1 , a sub 0 < zeta sub n $.
This can be done by starting with [$-1, 1$] and increasing the bounds if
not all the roots are between them (i.e. increase bounds until $w ( a sub 0 ) = 0$
and $w ( b sub 0 ) = n$).
Then divide this interval at its midpoint and check by means of the Sturm
sequence which of the two subintervals contains $zeta sub i$. The subinterval
which contains $zeta sub i$ is again divided, and so on. More precisely,
we form for $j = 0, 1, 2 ,...$
.EQ
u sub j == { ( a sub j + b sub j )} over 2 ,
.EN
.EQ
a sub {j + 1} == left { pile { { a sub j if w ( eta sub j ) >= n + 1 - i ,} above  { eta sub j if w ( eta sub j ) < n + 1 - i }}
.EN
.EQ
b sub {j + 1} == left { pile { { eta sub j if w ( eta sub j ) >= n + 1 - i ,} above  { b sub j if w ( eta sub j ) < n + 1 - i }}
.EN
The quantities $a sub j$ increase, and the quantities $b sub j$ decrease,
to the desired root $zeta sub i$. The convergence process is linear with
convergence rate 0.5. This method for determining the roots of the
characteristic polynomial is relatively slow, but very accurate. It also
has the feature, which makes it useful for our algorithm, that each root
can be determined independently of the others.
.br
Remember that we find this way only the eigenvalues of an irreducible
tridiagonal matrix - part of the given tridiagonal matrix. See next
sections for the algorithm to find the greatest eigenvalues of the whole
matrix.
.PP
There is a problem with the algorithm mentioned above. The polynomials in
(1), when evaluated at a certain point $x$, tend to grow large as the matrix
grows larger (for example, for a $N times N$ matrix, the last polynomial is
of degree $N$. This causes problems on computers, since numbers of them have
limited size, and numbers bigger than some number (about $10 sup 308$ on a Sun
computer) are treated like infinity, therefore no computations are possible).
So, given a point $x$, we seek a way of finding a different sequence of
values, $q sub 0 ,..., q sub n$ instead of $p sub 0 ( x ) ,..., p sub n ( x )$,
which have the same signs, but low absolute values. This can be done using two
methods I shall now describe:
.NH 2
The adaptive smoothing method
.PP
Let us look for a 
.I "smoothing sequence"
$sigma sub i$ of strictly positive
numbers, which we will apply in the following manner:
.EQ (2)
q sub 0 == p sub 0 ( x )
.EN
.EQ
q sub 1 == p sub 1 ( x ) sigma sub 1
.EN
.EQ
q sub 2 == p sub 2 ( x ) sigma sub 1 sigma sub 2
.EN
.EQ
q sub i == p sub i ( x ) sigma sub 1 sigma sub 2 ... sigma sub i , "    " i=2, 3, ..., n.
.EN
We want the smoothing sequence to be such that each $q sub i$ will have the
absolute value of 1, or 0. We shall then use the $q sub i$ sequence, instead
of the $p sub i ( x )$ sequence, which is possible since they have the same
signs, since $sigma sub i$ are strictly positive.
We start out by setting
$sigma sub 1$ to $1 over { | p sub 1 | }$. Then, as obvious from (2),
$q sub 1$ is $0$, $1$, or $-1$.
.\" $sigma sub 1$ should have a
.\".I nice
.\"(i.e.
.\"not extremely close to zero or extremely high) value, since we assume the
.\"tridiagonal matrix has 
.\".I nice
.\"elements, and so is $x$.
Now, let say that we have found $q sub 0 ,..., q sub { i - 1 }$,
and we are seeking $q sub i$. By (1), we know that
.EQ
p sub i ( x ) == ( alpha sub i - x) p sub {i - 1} (x) - {beta sub i} sup 2
p sub {i - 2} (x)
.EN
So that
.EQ I
q sub i mark == p sub i ( x ) sigma sub 1 sigma sub 2 ... sigma sub i
= sigma sub 1 sigma sub 1 ... sigma sub i cdot (( alpha sub i - x ) p sub {i - 1} ( x ) - {beta sub i} sup 2
p sub {i - 2} (x) )
.EN
.EQ I
lineup = sigma sub i cdot ( ( alpha sub i - x ) q sub { i - 1 } - { beta sub i }
sup 2 q sub { i - 2 } sigma sub {i - 1} )
.EN
We now define
.EQ
sigma sub i == 1 over { | ( alpha sub i - x ) q sub { i - 1 } - { beta sub i } sup 2 q sub { i - 2 } sigma sub { i - 1 } | }
.EN
Note that if the denominator has the value of zero, we set $sigma sub i$ to 1.
Now we have $q sub i$ which can be only $1$, $-1$ or $0$.
.\"$sigma sub i$ (which will next used for finding $q sub {i + 1}$) is the reciprocal
.\"of a non-zero number, and therefore cannot be infi
.\".I nice
.\"numbers, so it itself should be a
.\".I nice
.\"number.
.PP
This method of smoothing the sequence is very good, and if fact works for all
matrices I have tried. We shall now give an alternative smoothing method
which makes the whole Sturm sequence algorithm about 30% faster,
but unfortunately does not work good enough for
very large (about $400 times 400$) matrices. Note that one of these smoothing
methods
.I must
be used, since without them the Sturm sequence algorithm will fail even for
relatively small matrices.
.NH 2
The power smoothing method
.PP
This method also works by defining $q sub i$ based on $p sub i ( x )$
but does not adapt the smoothing value - it has a constant smoothing value
which is $sigma = 1 over { | x | }$, and the $q sub i$ are defined as follows:
.EQ
q sub 0 == p sub 0 ( x )
.EN
.EQ
q sub 1 == p sub 1 ( x ) sigma
.EN
.EQ
q sub 2 == p sub 2 ( x ) sigma sup 2
.EN
.EQ
q sub i == p sub i ( x ) sigma sup i , "    " i=3, ..., n.
.EN
This method works for modest sized matrices (less then about $400 times 400$)
since, ideally, a $n sup th$ degree polynomial evaluated at point $x$ divided
by $x sup n$ should be a
.I nice
number. We say ideally, since the polynomial coefficients themselves can grow
very large (remember: we never actually find those coefficients - we only find
the values at the point $x$ by recursion) rendering the
.I "power smoothing"
method useless. This method has problems with large matrices (about $400 times
400$), and may also not work very well with certain smaller matrices
.NH 2
Choosing the smoothing method
.PP
The routines I have written support both kinds of smoothing, and it is up to
the user to decide which one to use. I would recommend for users who want to
find eigenvalues or eigenvectors of a matrix of size $300 times 300$ or
smaller to try the power smoothing method, and if it fails (reports errors while
finding eigenvectors for the incorrect eigenvalues it found) try the
adaptive smoothing method. If you want to find eigenvalues and eigenvectors of
a larger matrix, or want to take no risks of finding incorrect eigenvalues,
then use the adaptive smoothing method. Remember that using the adaptive
smoothing method makes the Sturm sequence part of the algorithm about 30%
slower, but this is not real problem in large matrices, since the
Householder overhead dominates the running time.
.NH
Finding eigenvectors by inverse iteration
.PP
The following section discusses the problem of finding an eigenvector
corresponding to a given eigenvalue of an irreducible matrix. The
algorithm is outlined in [NRC], page 394.
.PP
We first have to note that because of the discussion in the previous section,
we have come to the conclusion that the characteristic polynomial of an
irreducible tridiagonal real symmetric matrix has only real roots, and what
is more important - all roots are simple. That saves us the trouble of
looking for more than one linearly independent vectors. Also, different
eigenvectors of such a matrix are automatically orthogonal, as they correspond
to different eigenvalues.
.PP
The basic idea behind inverse iteration is quite simple. Let $y$ be the
solution of the linear system
.EQ (1)
( A - tau I ) y = b
.EN
where b is a random vector and $tau$ is close to some eigenvalue of $A$. Then the
solution $y$ will be close to the eigenvector corresponding to $tau$.
The procedure can be iterated: replace $b$ by $y$ and solve for the new $y$,
which will be even closer to the true eigenvector.
Note that if $tau$ is too accurate, then equation (1) should not have a
solution, so we change $tau$ a little by adding some very small constant to
it, such as $10 sup -14$. Note that this causes the algorithm to have problems
if two eigenvalues are about this close to each other, but that should almost
never occur in practical uses (of course you can easily find a matrix which
causes this problem, but typical matrices will not cause this problem).
.PP
We can see why this works by expanding both $y$ and $b$ as linear combinations
of the eigenvectors $x sub j$ of A:
.EQ (2)
y = sum from {j} { alpha sub j x sub j }
"     "
b = sum from {j} { beta sub j x sub j }
.EN
Then (1) gives
.EQ (3)
sum from {j} { alpha sub j ( lambda sub j - tau ) x sub j }
= sum from {j} { beta sub j x sub j }
.EN
so that
.EQ (4)
alpha sub j = { beta sub j } over { lambda sub j - tau }
.EN
and
.EQ (5)
y = sum from {j} { { beta sub j x sub j } over { lambda sub j - tau } }
.EN
If $tau$ is close to $lambda sub n$, say, then provided $beta sub n$ is not
accidentally too small, $y$ will be approximately $x sub n$, up to normalization.
Moreover, each iteration of this procedure gives another power of $lambda sub j
- tau$ in the denominator of (5). Thus the convergence is rapid for well
separated eigenvalues (remember that we have already shown that all eigenvalues
are simple).
.PP
Suppose at the $i sup th$ stage of iteration we are solving the equation
.EQ (6)
( A - lambda sub i I ) y = x sub i
.EN
where $x sub i$ and $lambda sub i$ are our current guesses for the eigenvector 
and eigenvalue of interest (we shall not be updating $lambda sub i$, since it
is already accurate, but a variant of this algorithm also lets you improve
eigenvalues previously found).
.\" DROP THIS PARAGRAPH
.ig
The exact eigenvector and eigenvalue satisfy
.EQ (7)
A x = lambda x
.EN
so
.EQ (8)
( A - lambda sub i I ) x = ( lambda - lambda sub i ) x
.EN
..
Since $y$ of (6) is an improved approximation to $x$, we normalize it and set
.EQ (7)
x sub {i + 1} = y over { | y | }
.EN
.PP
These formulas look simple enough, but in practice, the implementation is
quite tricky
.\", and in fact, in the algorithm I wrote there are still some bugs.
.PP
We begin by choosing a random normalized vector, as the initial guess for
the eigenvector, $x sub 0$. We then solve (6). If the solving algorithm fails
(I will not discuss here the algorithm to solve a tridiagonal system of
equations, since it is a simple Gaussian elimination for tridiagonal system)
then we
choose a new random vector and start again. If the solution of (6) succeeded,
we check if $| x sub 1 - x sub 0 | < epsilon$, where $epsilon$ is some
tolerance, which I have chosen in my program to be $10 sup { - 15 }$. If it
is indeed smaller, then we have found our eigenvector $x sub 1$. If not
we keep iterating. If after some number of iterations, say 10 iterations,
$| x sub {i + 1} - x sub i |$ has not decreased enough, then we begin again
with a new random vector.
If, after a new random vector has been chosen, let say, 3 times, we still did
not find an eigenvector, we regretfully have to say that the algorithm failed.
This should not happen, but, unfortunately, I still have some problems with
giant matrices (about $800x800$).
(Of course, if you are unlucky enough to get 3 consecutive bad random vectors,
then that's another problem. However, this seems very unlikely.)
.PP
This algorithm the complexity of $O ( n sup 3 )$, and it is much less
efficient than the QL method (cf. [NRC]). However, when less then about
25 percent of the eigenvectors are required, this algorithm should be more
efficient.

.NH
Putting the algorithms together
.PP
Now that I have explained all those algorithms, it's time to put them together
for our final algorithm. Given $A$ - a $n times n$ real symmetric matrix, we want to find it's
$s$ biggest eigenvalues, and respective eigenvectors.
.PP
We begin by using the Householder algorithm to transform A into a tridiagonal
matrix. Call the tridiagonal matrix $T$ and the orthogonal transformation
matrix $P$.
We can't proceed with the Sturm sequence algorithm, and the inverse iteration
algorithms because they only work on an irreducible tridiagonal matrix.
So we break $T$ into square tridiagonal blocks $T sub i , ~ i = 1 ,..., m,$
which are irreducible.
Note that any eigenvalue of a block $T sub i$ is also an eigenvalue of $T$,
and vice versa - each eigenvalue of $T$ is an eigenvalue of some $T sub i$,
and if $lambda$ is non-simple eigenvalue of degree $j$ of $T$ then there are exactly
$j$ blocks with a (simple) eigenvalue $lambda$.
This is obvious from the block structure of $T$ (the characteristic polynomial
of $T$
is the product of the characteristic polynomials of $T sub i$). Also, if we
have found an eigenvector of $T sub i$, then to find a corresponding
eigenvector of $T$ all we have to do is add zeros to the vector to make it
line up with the appropriate block $T sub i$ in the matrix multiplication
$T v = lambda v$.
.PP
Now we make an array of size $s$ to remember the
eigenvalues. First we find the biggest eigenvalue of each irreducible
block $T sub i$ with the Sturm sequence algorithm from section 4,
and sort them into the array. Now, we find the second biggest
eigenvalue of each block. We merge them into the array. If the array was filled
with eigenvalues and $T sub i$'s second eigenvalue was smaller then the smallest
eigenvalue in the array, then we can stop looking at $T sub i$ in the next iteration, since it's next eigenvalues are even smaller. We also drop a $T sub i$
if we have finished getting all it's eigenvalues.
We repeat this process, until we ran out of $T sub i$'s.
We are left with the $s$ biggest eigenvalues of $T$ in the array. Now, for
each of those eigenvalues we find the eigenvector corresponding to it, in the
appropriate irreducible block $T sub i$. We normalize this vector.
We then add zeros to make it an eigenvector
of $T$.
We still have to see that the eigenvectors we get this way are orthogonal, however
this is obvious: Two eigenvectors which were gotten from the same irreducible
block correspond to different eigenvalues, and therefore are automatically
orthogonal. Two eigenvectors which were gotten from different irreducible
blocks are obviously orthogonal since their non-zero elements are at a
different place.
.PP
Now that we have the $s$ biggest eigenvalues of $T$, and their corresponding
orthonormal eigenvectors, all we have to do is to multiply them by
the orthogonal transformation matrix $P$. We then finally have the $s$
biggest eigenvalues of $A$, and their corresponding orthonormal eigenvectors.
.NH
Performance of the Algorithm
.PP
One of the goals of this project was to make the algorithm presented here as
fast as possible. If time was not important, then one can use one of the
many generic algorithms floating around which find all the eigenvalues and
eigenvectors, such as the routines implemented in Matlab\(dg:
.FS
\(dg Matlab Copyright (c) The MathWorks, Inc. 1984-1991
.FE
the $eig$
command of Matlab finds all the eigenvalues and eigenvectors of any matrix.
While this may be useful in some cases, it is not needed in our case, and
as we shall see, is much slower than our algorithm when only a part of the
eigenvalues and eigenvectors are sought. In this section we shall also see
examples of the time used by our algorithm as function of $n$ - the size
of the given matrix and $s$ - the number of eigenvalues and eigenvectors
wanted.
.PP
All of the following examples were run on tx.technion.ac.il, which is a
Sun Sparcstation, with 690MP CPU (four processors), with the SunOS 4.1.2
operation system (Note that although the computer has four processors,
the current algorithm has no support for parallel execution. An enhancement
to the algorithm could be to calculate eigenvectors in parallel).
.PP
I have written a program incorporating the algorithm described in this report,
and ran an example in which I find eigenvectors and eigenvalues of
some fixed $200 times 200$ random matrix (i.e. a $200 times 200$ random matrix
was chosen, and remained fixed through all the following tests). The following
table shows the time it took this program to run, and the time it took matlab
to run on the same matrix. The program 'main3' in the following table is the
program which uses the algorithm described in this report.
.TS
center box ;
cb s s
cb | cb | cb
c | n | n 
^ | n | n
^ | n | n
^ | n | n
^ | n | n
^ | n | n
c | n | n .
Table 1 - a $bold {200 times 200}$ matrix
_
Program	Eigs wanted	CPU seconds
_
main3	200	21.80
	50	11.10
	30	9.50
	20	8.90
	10	8.15
	0	7.30
_
Matlab	200	22.00
.TE
One can easily see from table 1 that main3 is faster than matlab, even when
all 200 eigenvalues and eigenvectors are wanted. This is because the algorithm
employed by Matlab does not use the fact that the matrix is symmetric.
We can also see from this table how decreasing the number of wanted
eigenvalues and eigenvectors decreases the time the program takes to run.
Asking for only 50 out of the 200 eigenvalues and eigenvectors decreases the
run time to half. Also, by looking at this table we can see that the Householder
routine took 7.3 seconds to run (because finding 0 eigenvalues means only
do the Householder routine), and every eigenvalue and eigenvector wanted took
about 0.072 second to find.
.PP
The following table shows how does the size of
the matrix influence the running time of the algorithm. In the table, all
times are in CPU seconds, i.e. the time the computer really worked on the program.
The 'Sturm&Inverse + Convert' heading specifies the average time it took to find
$ {roman "one"} under$ eigenvalue and corresponding eigenvector of the $n times n$ random matrix.
The 'Sturm&Inverse' time includes the Sturm algorithm to find eigenvalues
and the inverse iteration algorithm to find the corresponding eigenvector,
and the 'Convert' time specifies the time of the the matrix multiplication
needed to convert the eigenvector to the one of the original matrix.
.TS
center box delim($$) ;
cb s s
cb | cb | cb
n | n | n .
Table 2 - run time of parts of the algorithm (in CPU seconds)
_
Matrix size	Householder	Sturm&Inverse + Convert
_
10	0.00	$0.000+0.000=0.000$
20	0.00	$0.004+0.001=0.005$
30	0.02	$0.006+0.001=0.007$
40	0.05	$0.007+0.003=0.010$
50	0.10	$0.008+0.002=0.010$
60	0.18	$0.009+0.004=0.013$
80	0.37	$0.013+0.005=0.018$
100	0.77	$0.015+0.010=0.025$
150	2.83	$0.022+0.024=0.046$
200	7.02	$0.030+0.042=0.072$
300	29.67	$0.045+0.097=0.142$
400	111.88	$0.061+0.172=0.233$
500	167.75	$0.079+0.270=0.349$
600	344.83	$0.101+0.384=0.485$
700	555.98	$0.126+0.525=0.651$
800	989.75	$0.145+0.688=0.833$
.TE
As can be seen quite clearly from this table (for example, compare the time
for a $800 times 800$ matrix to the time of a $400 times 400$ matrix),
the complexity of the Householder
algorithm is nearly $O ( n sup 3 )$ for $n -> inf$, like I mentioned before.
This is really bad
for extremely large matrices, but unfortunately it is the best algorithm
available ([NCR] says that the Householder algorithm together with a QL
algorithm with implicit shifts is the most efficient known technique for finding
.I all
the eigenvalues and eigenvectors of a real symmetric matrix).
From table 2 we can also see that the Sturm sequence and inverse iteration
algorithms combined are approximately $O ( n )$. The conversion of an
eigenvector of the tridiagonal matrix to the one of the original matrix
looks like a simple operation, but unfortunately, as can be seen in table 2,
it has the complexity of $O ( n sup 2 )$, so for large $n$, the conversion
takes much more time then the Sturm and inverse iteration algorithms!
This is because the conversion works with the whole matrix, which is $n times n$,
as opposed to working with a tridiagonal matrix which has only $2 n - 1$ elements.
.PP
We have come to the conclusion, that the whole algorithm is $O ( n sup 3 )$
when $n$ is large enough. Because I said the most efficient algorithm to
find eigenvalues and eigenvectors of real symmetric matrices uses the
Householder algorithm, then even the most efficient algorithm must be
$O ( n sup 3 )$. Knowing that we can't improve the Householder algorithm,
we look for ways to increase the speed of the second part of the algorithm
(Sturm sequence bisection, inverse iteration, and matrix multiplication).
Suppose, for example, that we have a $n times n$ matrix, and we want to
find half its eigenvalues and eigenvectors. Then, we use a $O ( n sup 2 )$
algorithm $n over 2$ times, so overall we get a $O ( n sup 3 )$ complexity.
Improving small details of the algorithm (see next section for planned
improvements) will not make it much faster. We can really make that part
of the algorithm much faster if we are working on a computer with several
CPU's, such as a supercomputer. As we previously saw, an important benefit
of this algorithm is that each eigenvalue and eigenvector can be found
independently of the others. So we can use different processors to find
different eigenvalues and their corresponding eigenvectors. For example,
let us say we want to find the 200 biggest eigenvalues of the $800 times 800$
matrix mentioned in table 2. Using the regular sequential algorithm will
take us $989.75 + 200 times 0.833 = 1156.35$ seconds. Using 100 CPU's, each
finding two eigenvalues and the two corresponding eigenvectors, will take us
only $989.75 + 2*0.833 = 991.416$
seconds, so we saved $164.934$ seconds. If the matrix is bigger, or we want
more eigenvalues and eigenvectors, then the absolute time difference will be
even greater. This way we decrease the running time of the second part of
the algorithm to $1 over 100$ of the original time (don't forget that we still
have the $O ( n sup 3 )$ Householder overhead, which isn't decreased by using
a computer with several processors. This is because the Householder algorithm
is sequential by nature, and each step of it requires the results of the
previous step).
.NH
Summary
.PP
As can be seen from this report, finding the biggest eigenvalues and
corresponding eigenvectors can be a tough job. It is even harder trying to
code the algorithm with a computer language, as you run into many unexpected
problems. The most problematic algorithm to code in a machine with finite
accuracy are the inverse iteration algorithm, together with it's algorithm
for solving a tridiagonal system of equations, and the Sturm sequence
algorithm. As I said before, this algorithm
is not perfect yet, and runs into trouble in some cases of giant matrices.
.PP
Another thing that can be done to improve the final algorithm is to improve
the convergence speed of the Sturm sequence algorithm. One way to do this
is instead of always using the bounds [$a sub 0 , b sub 0$] which are good
for all the eigenvalues, use more accurate bounds: when checking the place
of one eigenvalue we automatically improve our guess on other eigenvalues'
whereabouts. The other, more dramatic, change would be to completely discard
the bisection method, and instead use a newton-related method, which is
perhaps less accurate, but faster then the bisection method.

.NH
Acknowledgments
.PP
The project described in this report was supported by the US-Israel Binational
Science Foundation, under supervision of Prof. Koby Rubinstein, Department of
Mathematics, Technion - Israel Institute of Technology.
.PP
I would also like to thank my father, Zvi Har'El, for giving me very useful
ideas.
.bp
.NH
References.
.IP [NRC]
Press, H. W. & Flannery, B. P. & Teukolsky A. T. & Vetterling W. T.,
.I "Numerical Recipes in C, The Art of Scientific Computing" ,
Chapter 11.
Cambridge University Press
.IP [KL1]
Sirovich, L. & Kirby, M.,
.I "Low-dimensional procedure for the characterization of human faces" ,
Journal of the Optical Society of America, March 1987, Vol 4, No. 3, Page 519.
.IP [KL2]
Sirovich, L. & Kirby, M.,
.I "Application of the Karhunen-Loeve Procedure for the Characterization of Human Faces" ,
IEEE Transactions on Pattern Analysis and Machine Intelligence, Vol. 12, No 1, January 1990, Page 103.
.IP [INA]
Stoer, J. & Bulirsch, R.,
.I "Introduction to Numerical Analysis" ,
Chapter 5.6.
Springer-Verlag New York, 1980.
