#include "eigen.h"

householder(a,n,t)
	matrix a;	/* input: NxN regular matrix.
			   on output it is replaced by the orthogonal
			   transformation matrix */
	int n;
	matrix t;	/* output: NxN tridiagonal matrix */
{
	int l,k,j,i;
	scalar scale,hh,h,g,f;
    
	for (i=n-1;i>=1;i--) {
		l=i-1;
		h=scale=0.0;
		if (l > 0) {
			for (k=0;k<=l;k++)
				scale += ABS(a[i][k]);
			if (scale == 0.0)
				t[1][l]=a[i][l];
			else {
				for (k=0;k<=l;k++) {
					a[i][k] /= scale;
					h += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = f>0 ? -sqrt(h) : sqrt(h);
				t[1][l]=scale*g;
				h -= f*g;
				a[i][l]=f-g;
				f=0.0;
				for (j=0;j<=l;j++) {
				/* Next statement can be omitted if eigenvectors not wanted */
					a[j][i]=a[i][j]/h;
					g=0.0;
					for (k=0;k<=j;k++)
						g += a[j][k]*a[i][k];
					for (k=j+1;k<=l;k++)
						g += a[k][j]*a[i][k];
					t[1][j-1]=g/h;
					f += t[1][j-1]*a[i][j];
				}
				hh=f/(h+h);
				for (j=0;j<=l;j++) {
					f=a[i][j];
					g=t[1][j-1]-=hh*f;
					for (k=0;k<=j;k++)
						a[j][k] -= (f*t[1][k-1]+g*a[i][k]);
				}
			}
		} else
			t[1][l]=a[i][l];
		t[0][i]=h;
	}
	/* Next statement can be omitted if eigenvectors not wanted */
	t[0][0]=0.0;
	/*t[1][-1]=0.0;*/
	/* Contents of this loop can be omitted if eigenvectors not
			wanted except for statement t[0][i]=a[i][i]; */
	for (i=0;i<n;i++) {
		l=i-1;
		if (t[0][i]) {
			for (j=0;j<=l;j++) {
				g=0.0;
				for (k=0;k<=l;k++)
					g += a[i][k]*a[k][j];
				for (k=0;k<=l;k++)
					a[k][j] -= g*a[k][i];
			}
		}
		t[0][i]=a[i][i];
		a[i][i]=1.0;
		for (j=0;j<=l;j++) a[j][i]=a[i][j]=0.0;
	}
}
